"""Test the validators module."""
import pytest
from jsonschema.exceptions import ValidationError

from diameter_synthesis import validators


def test_defaults(config, model_params):
    """Test the validation of JSON schemas."""
    mtype = "L5_TPC:A"
    model = "generic"

    validators.validate_model_params(model_params[model][mtype])
    validators.validate_config(config[model])

    config[model]["features"] = {"FEATURE 1": {"apical": 0.2}}
    validators.validate_config(config[model])

    config[model]["features"] = {"FEATURE 2": {"UNKNOWN": 0.2}}
    with pytest.raises(ValidationError):
        validators.validate_config(config[model])



"""Tools to validate the input parameters."""
import json

from jsonschema import validate
from pkg_resources import resource_stream

MODEL_SCHEMA = json.load(resource_stream("diameter_synthesis", "schemas/model_params.json"))
CONFIG_SCHEMA = json.load(resource_stream("diameter_synthesis", "schemas/config.json"))


def validate_model_params(data):
    """Validate model parameter dictionary."""
    validate(data, MODEL_SCHEMA)


def validate_config(data):
    """Validate config dictionary."""
    validate(data, CONFIG_SCHEMA)



"""Click app."""
import logging
import os
from pathlib import Path

import click
import morphio
from tqdm import tqdm

from .utils import create_morphologies_dict

morphio.set_maximum_warnings(0)
L = logging.getLogger(__name__)
logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))


# pylint: disable=import-outside-toplevel,redefined-outer-name


@click.group()
def cli():
    """Cli to learn and generate diameters of neurons."""


@cli.command("run_models")
@click.argument("config_file", type=click.Path(exists=True))
@click.option("--plot", is_flag=True)
@click.option("--ext", default=".png")
def run_models(config_file, plot=False, ext=".png"):
    """Run the model extraction from config file."""
    from .main import run_models

    run_models(config_file, plot=plot, ext=ext)


@cli.command("run_diameters")
@click.argument("config_file", type=click.Path(exists=True))
@click.argument("models_params_file", type=click.Path(exists=True))
def run_diameters(config_file, models_params_file):
    """Build new diameters from config file and diameter model."""
    from .main import run_diameters

    run_diameters(config_file, models_params_file)


@cli.command("plot_diff")
@click.argument("original_folder", type=click.Path(exists=True))
@click.argument("diametrized_folder", type=click.Path(exists=True))
@click.argument("plot_folder", type=click.Path())
@click.option("-n", "--ncells", help="max number of cells to plot")
@click.option("-e", "--ext", help="figures extention")
def plot_diff(original_folder, diametrized_folder, plot_folder, ncells=None, ext=".png"):
    """Plot original and new neurons as well as their differences."""
    from .plotting import plot_diameter_diff

    neurite_types = ["basal", "apical", "axon"]
    import neurom as nm

    morphologies_dict = create_morphologies_dict(original_folder)

    if not Path(plot_folder).exists():
        os.mkdir(plot_folder)

    for mtype in morphologies_dict:
        L.info("Plot mtype %s", mtype)

        plot_folder_mtype = Path(plot_folder) / mtype
        if not Path(plot_folder_mtype).exists:
            os.mkdir(plot_folder_mtype)

        if ncells is not None:
            neurons = morphologies_dict[mtype][: int(ncells)]
        else:
            neurons = morphologies_dict[mtype]

        for neuron in tqdm(neurons):
            neuron_new = nm.load_morphology(Path(diametrized_folder) / neuron.name)

            plot_diameter_diff(neuron, neuron_new, neurite_types, plot_folder_mtype, ext=ext)


@cli.command("run_analysis")
@click.option("--orig-path", help="Path to original cells", required=True)
@click.option("--diam-path", help="Path to diametrized cells", required=True)
@click.option("--mtypes-file", help="Path to mtypes file", required=False)
@click.option("-o", "--out-dir", help="Directory to output the analysis results", required=True)
@click.option("--cumulative", help="Cumulative distribution plots", is_flag=True)
@click.option("--individual", help="Output a plot for each neuron", is_flag=True)
@click.option("--violin", help="Violin distribution plots", is_flag=True)
@click.option("-e", "--ext", help="Figures extention")
def run_analysis(
    orig_path, diam_path, out_dir, cumulative, individual, violin, mtypes_file=None, ext=".png"
):
    """Produce figures for validation/analysis."""
    if not cumulative and not violin:
        raise ValueError("Should at least set one of '--cumulative' or '--violin' to True.")
    if cumulative:
        from .plotting import cumulative_analysis

        cumulative_analysis(
            orig_path,
            diam_path,
            Path(out_dir) / "basal",
            individual,
            mtypes_file=mtypes_file,
            neurite_types=["basal"],
            ext=ext,
        )
        cumulative_analysis(
            orig_path,
            diam_path,
            Path(out_dir) / "axon",
            individual,
            mtypes_file=mtypes_file,
            neurite_types=["axon"],
            ext=ext,
        )
        cumulative_analysis(
            orig_path,
            diam_path,
            Path(out_dir) / "apical",
            individual,
            mtypes_file=mtypes_file,
            neurite_types=["apical"],
            ext=ext,
        )
    if violin:
        from .plotting import violin_analysis

        violin_analysis(orig_path, diam_path, out_dir, mtypes_file=mtypes_file)



"""Test the distribution_fitting module."""
import logging

import pytest

from diameter_synthesis import distribution_fitting
from diameter_synthesis.exception import DiameterSynthesisError


def test_fit_distribution(config, caplog):
    """Test the fit_distribution function."""
    model = "generic"
    all_data = [
        [0.75, 0.25],
        [0.7, 0.5],
        [0.8, 0.001],
        [0.99, 0.033],
        [0.428, 0.254],
        [0.808, 0.006],
        [0.397, 0.537],
        [0.72, 0.223],
        [0.243, 0.02],
        [0.342, 0.244],
    ]
    data_1d = [i[0] for i in all_data]
    config[model]["neurite_type"] = "basal"

    # Test expon_rev distribution
    res_expon_rev = distribution_fitting.fit_distribution(
        data_1d, "expon_rev", extra_params=config[model]
    )
    assert res_expon_rev == pytest.approx(
        {"loc": -0.99, "scale": 0.3722, "min": 0.28755, "max": 0.99, "num_value": 10.0}
    )

    # Test exponnorm distribution
    res_exponnorm = distribution_fitting.fit_distribution(
        data_1d, "exponnorm", extra_params=config[model]
    )

    assert res_exponnorm == pytest.approx(
        {
            "a": 0.3,
            "loc": 0.550945,
            "scale": 0.224617,
            "min": 0.28755,
            "max": 0.9081,
            "num_value": 10,
        }
    )

    # Test gamma distribution
    res_gamma = distribution_fitting.fit_distribution(data_1d, "gamma", extra_params=config[model])
    assert res_gamma == pytest.approx(
        {"a": 4.0, "loc": -1e-09, "scale": 0.15445, "min": 0.28755, "max": 0.9081, "num_value": 10}
    )

    # Test unknown distribution
    with pytest.raises(DiameterSynthesisError):
        distribution_fitting.fit_distribution(all_data, "UNKNOWN", extra_params=config[model])

    # Test unknown attribute name
    with pytest.raises(DiameterSynthesisError):
        distribution_fitting.fit_distribution(
            data_1d, "expon_rev", attribute_name="UNKNOWN", extra_params=config[model]
        )

    # Test with empty input data
    caplog.clear()
    caplog.set_level(logging.WARNING)
    config[model]["name"] = "sibling_ratios"
    res_no_data = distribution_fitting.fit_distribution([], "expon_rev", extra_params=config[model])

    assert res_no_data == {
        "a": 0.0,
        "loc": 0.0,
        "scale": 0.0,
        "min": 0.0,
        "max": 0.1,
        "num_value": 0,
    }

    assert len(caplog.record_tuples) == 1
    for i in caplog.record_tuples:
        module, level, entry = i
        assert module == "diameter_synthesis.distribution_fitting"
        assert level == 30
        assert entry == "Not enough data to fit distribution sibling_ratios with 0 points"

    # Test A_MIN and A_MAX
    A_MIN = distribution_fitting.A_MIN
    A_MAX = distribution_fitting.A_MAX
    try:
        distribution_fitting.A_MIN = 999
        distribution_fitting.A_MAX = 999

        # Test exponnorm distribution
        res_exponnorm_MIN_MAX = distribution_fitting.fit_distribution(
            data_1d, "exponnorm", extra_params=config[model]
        )
        assert res_exponnorm_MIN_MAX == pytest.approx(
            {
                "a": 999,
                "loc": 0.24197457,
                "scale": 0.00037496929,
                "min": 0.28755,
                "max": 0.9081,
                "num_value": 10,
            }
        )

        # Test gamma distribution
        res_gamma_MIN_MAX = distribution_fitting.fit_distribution(
            data_1d, "gamma", extra_params=config[model]
        )
        assert res_gamma_MIN_MAX == pytest.approx(
            {
                "a": 999,
                "loc": -1e-09,
                "scale": 0.000618418,
                "min": 0.28755,
                "max": 0.9081,
                "num_value": 10,
            }
        )
    finally:
        distribution_fitting.A_MIN = A_MIN
        distribution_fitting.A_MAX = A_MAX

    # Test with attribute name
    config[model]["features"] = {
        "TEST FEATURE": {
            "basal": 0.5,
        }
    }
    res_expon_rev_feature = distribution_fitting.fit_distribution(
        all_data, "expon_rev", attribute_name="TEST FEATURE", extra_params=config[model]
    )
    assert res_expon_rev_feature == pytest.approx(
        {"loc": -0.99, "scale": 0.354875, "min": 0.27765, "max": 0.99, "num_value": 8.0}
    )


def test_sample_distribution():
    """Test the sample_distribution function."""
    model = {
        "params": {
            "a": 1.5,
            "loc": 1,
            "max": 4,
            "min": 0.5,
            "num_value": 500,
            "scale": 0.5,
        },
        "sequential": None,
    }

    # Test expon_rev distribution
    model["distribution"] = "expon_rev"
    model["params"]["loc"] = -1
    res_expon_rev = distribution_fitting.sample_distribution(model)
    assert res_expon_rev == pytest.approx(0.602062745918445)

    # Test exponnorm distribution
    model["distribution"] = "exponnorm"
    model["loc"] = 1
    res_exponnorm = distribution_fitting.sample_distribution(model)
    assert res_exponnorm == pytest.approx(0.5337329679223535)

    # Test gamma distribution
    model["distribution"] = "gamma"
    model["loc"] = 1
    res_gamma = distribution_fitting.sample_distribution(model)
    assert res_gamma == pytest.approx(0.5690896450847811)

    # Test unknown distribution
    model["distribution"] = "UNKNOWN"
    model["loc"] = 1
    with pytest.raises(DiameterSynthesisError):
        distribution_fitting.sample_distribution(model)

    # Test truncate fail
    model["distribution"] = "expon_rev"
    model["params"]["loc"] = 999
    with pytest.raises(DiameterSynthesisError):
        distribution_fitting.sample_distribution(model)


def test_evaluate_distribution():
    """Test the evaluate_distribution function."""
    # Test expon_rev distribution
    res_expon_rev = distribution_fitting.evaluate_distribution(
        -1, "expon_rev", {"loc": 1, "scale": 10}
    )
    assert res_expon_rev == pytest.approx(0.1)

    # Test exponnorm distribution
    res_expon_rev = distribution_fitting.evaluate_distribution(
        1, "exponnorm", {"a": 1, "loc": 1, "scale": 10}
    )
    assert res_expon_rev == pytest.approx(0.026157829186512344)

    # Test exponnorm distribution
    res_expon_rev = distribution_fitting.evaluate_distribution(
        1, "gamma", {"a": 1, "loc": 1, "scale": 10}
    )
    assert res_expon_rev == pytest.approx(0.1)

    # Test unknown distribution
    with pytest.raises(DiameterSynthesisError):
        distribution_fitting.evaluate_distribution(0, "UNKNOWN", {})



"""Construct diameter models from cells."""
import logging
from functools import partial

from neurom import NeuriteType

import diameter_synthesis.morph_functions as morph_funcs
from diameter_synthesis.distribution_fitting import fit_distribution
from diameter_synthesis.exception import DiameterSynthesisError

L = logging.getLogger(__name__)

STR_TO_NEUROM_TYPES = {
    "apical": NeuriteType.apical_dendrite,
    "basal": NeuriteType.basal_dendrite,
    "axon": NeuriteType.axon,
}


def _get_model_builder(config):
    """Get the diameter model builder.

    Args:
        config (dict): configuration dictionary

    Returns:
        dict: dictionary of models
    """
    all_models = {}
    for model in config["models"]:
        if model == "generic":
            distribution_types = {
                "sibling_ratios": ["expon_rev", None],
                "diameter_power_relation": ["exponnorm", None],
                "terminal_diameters": ["exponnorm", None],
                "trunk_diameters": ["exponnorm", None],
                "trunk_diameters_method": "last",
                "tapers": ["expon_rev", None],
            }
        elif model == "astrocyte":
            distribution_types = {
                "sibling_ratios": ["expon_rev", None],
                "diameter_power_relation": ["exponnorm", None],
                "terminal_diameters": ["exponnorm", None],
                "trunk_diameters_method": "mean",
                "trunk_diameters": ["exponnorm", None],
                "tapers": ["exponnorm", None],
            }
        else:
            raise DiameterSynthesisError(f"model not understood {model}")

        all_models[model] = partial(build_single_model, distribution_types)
    return all_models


def build(morphologies, config, with_data=False):
    """Build a model from a set of morphologies and a config file.

    Args:
        morphologies (dict): dictionary of morphologies
        config (dict): configuration dictionary

    Returns:
        dict: parameter of the models (possibly the data extracted)
    """
    all_models = _get_model_builder(config)
    models_params, models_data = all_models[config["models"][0]](morphologies, config)
    if with_data:
        return models_params, models_data
    return models_params


def build_single_model(sampling_model, morphologies, config):
    """Build a single model.

    Args:
        sampling_model (dict): parameters for model building
        morphologies (dict): dictionary of morphologies
        config (dict): configuration dictionary

    Returns:
        dict: parameter of the models data extracted)
    """
    all_data = extract_parameters(sampling_model, morphologies, config)
    all_models = fit_all_models(all_data, sampling_model, config)
    return all_models, all_data


def extract_parameters(sampling_model, morphologies, config):
    """Extract parameters from neurites to then use for model building..

    Args:
        sampling_model (dict): parameters for model building
        morphologies (dict): dictionary of morphologies
        config (dict): configuration dictionary

    Returns:
        dict: data extracted
    """
    all_data = {
        "sibling_ratios": {},
        "diameter_power_relation": {},
        "terminal_diameters": {},
        "trunk_diameters": {},
        "tapers": {},
    }

    for neurite_type in config["neurite_types"]:
        all_data["sibling_ratios"][neurite_type] = []
        all_data["diameter_power_relation"][neurite_type] = []
        all_data["terminal_diameters"][neurite_type] = []
        all_data["trunk_diameters"][neurite_type] = []
        all_data["tapers"][neurite_type] = []

        for neuron in morphologies:
            for neurite in neuron.neurites:
                if neurite.type == STR_TO_NEUROM_TYPES[neurite_type]:
                    all_data["sibling_ratios"][neurite_type] += morph_funcs.compute_sibling_ratios(
                        neurite, attribute_name=sampling_model["sibling_ratios"][1]
                    )
                    all_data["diameter_power_relation"][
                        neurite_type
                    ] += morph_funcs.compute_diameter_power_relation(
                        neurite,
                        attribute_name=sampling_model["diameter_power_relation"][1],
                    )
                    all_data["terminal_diameters"][neurite_type] += morph_funcs.terminal_diameters(
                        neurite,
                        threshold=config["terminal_threshold"],
                        attribute_name=sampling_model["terminal_diameters"][1],
                    )
                    all_data["trunk_diameters"][neurite_type] += morph_funcs.trunk_diameter(
                        neurite,
                        attribute_name=sampling_model["trunk_diameters"][1],
                        method=sampling_model["trunk_diameters_method"],
                    )
                    all_data["tapers"][neurite_type] += morph_funcs.taper(
                        neurite,
                        params=config["taper"],
                        attribute_name=sampling_model["tapers"][1],
                    )
    return all_data


def fit_all_models(all_data, sampling_model, config):
    """Fit the parmaeters to get models.

    Args:
        all_data (dict): parameters to fit
        sampling_model (dict): parameters for model building
        config (dict): configuration dictionary

    Returns:
        dict: models
    """
    all_models = {
        "sibling_ratios": {},
        "diameter_power_relation": {},
        "terminal_diameters": {},
        "trunk_diameters": {},
        "tapers": {},
    }

    extra_params = config.copy()
    for neurite_type in config["neurite_types"]:

        extra_params["neurite_type"] = neurite_type
        extra_params["name"] = "sibling_ratios"
        all_models["sibling_ratios"][neurite_type] = fit_model(
            sampling_model["sibling_ratios"],
            all_data["sibling_ratios"][neurite_type],
            extra_params,
        )

        extra_params["name"] = "diameter_power_relation"
        all_models["diameter_power_relation"][neurite_type] = fit_model(
            sampling_model["diameter_power_relation"],
            all_data["diameter_power_relation"][neurite_type],
            extra_params,
        )

        extra_params["name"] = "terminal_diameter"
        all_models["terminal_diameters"][neurite_type] = fit_model(
            sampling_model["terminal_diameters"],
            all_data["terminal_diameters"][neurite_type],
            extra_params,
        )

        extra_params["name"] = "trunk_diameters"
        all_models["trunk_diameters"][neurite_type] = fit_model(
            sampling_model["trunk_diameters"],
            all_data["trunk_diameters"][neurite_type],
            extra_params,
        )

        extra_params["name"] = "tapers"
        all_models["tapers"][neurite_type] = fit_model(
            sampling_model["tapers"], all_data["tapers"][neurite_type], extra_params
        )

    return all_models


def fit_model(sampling_model, data, extra_params):
    """Fit a single parameter.

    Args:
        sampling_model (dict): model parameters
        data (array): parameter to fit
        extra_aprams (dict): additional parameters

    Returns:
        dict: fit parameters
    """
    if len(data) == 0:
        return {}
    return {
        "distribution": sampling_model[0],
        "sequential": sampling_model[1],
        "params": fit_distribution(
            data,
            sampling_model[0],
            attribute_name=sampling_model[1],
            extra_params=extra_params,
        ),
    }



"""Test the build_diameters module."""
import logging

import morphio.mut
import numpy as np
import pytest
from morphio import PointLevel
from morphio import SectionType

from diameter_synthesis import build_diameters
from diameter_synthesis.exception import DiameterSynthesisError

from .testing_tools import _compare_diameters


def test_build_simple(config, model_params, small_morph):
    """Test the main build function to diametrize a neuron."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"

    build_diameters.build(small_morph, model_params[model][mtype], neurite_types, config[model])

    expected = morphio.mut.Morphology(small_morph)
    diameters = {
        0: [1.12723756, 0.8501913, 0.67725074],  # basal
        1: [1.88172221, 1.60697436, 1.33222628],  # basal
        2: [0.7807585, 0.5290243, 0.52312386],  # basal
        3: [0.80801165, 0.62176675, 0.59818721],  # basal
        4: [3.19154882, 2.68855309, 2.1855576],  # apical
        5: [1.23321462, 0.85168982, 0.63268346],  # apical
        6: [1.1493628, 0.83900893, 0.72536111],  # apical
        7: [0.52783465, 0.44990396, 0.42342067],  # apical
        8: [0.51564646, 0.40626049, 0.37496084],  # apical
        9: [0.36634988, 0.40507823, 0.45199862],  # apical
        10: [0.36634988, 0.40620297, 0.47370234],  # apical
        11: [0.41420197, 0.42738733, 0.46979269],  # apical
        12: [0.41356379, 0.38290364, 0.389568],  # apical
        13: [0.2, 0.2, 0.2],  # axon
        14: [0.2, 0.2, 0.2],  # axon
        15: [0.2, 0.2, 0.2],  # basal
        16: [0.2, 0.2, 0.2],  # basal
    }
    for section_id, section in expected.sections.items():
        section.diameters = diameters[section_id]

    _compare_diameters(expected, small_morph)


def test_build_simple_with_apical_sections(config, model_params, small_morph):
    """Test the main build function to diametrize a neuron with a known apical section."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"
    model_params[model][mtype]["apical_point_sec_ids"] = [6]

    build_diameters.build(small_morph, model_params[model][mtype], neurite_types, config[model])

    expected = morphio.mut.Morphology(small_morph)
    diameters = {
        0: [1.57616556, 1.10349214, 0.71787846],  # basal
        1: [1.9246546, 1.59596884, 1.29954779],  # basal
        2: [0.8029623, 0.5610494, 0.53921282],  # basal
        3: [0.92826194, 0.69051898, 0.61900127],  # basal
        4: [3.54046178, 3.02864218, 2.51682234],  # apical
        5: [0.49307173, 0.38693029, 0.38693029],  # apical
        6: [2.51682234, 2.15803671, 1.8276546],  # apical
        7: [0.90557307, 0.65467209, 0.48454937],  # apical
        8: [1.32304311, 1.06992745, 0.8688401],  # apical
        9: [0.71307242, 0.5796622, 0.47997007],  # apical
        10: [0.68834245, 0.61472285, 0.61060423],  # apical
        11: [0.51838523, 0.38276276, 0.38276276],  # apical
        12: [0.46621996, 0.42117757, 0.44140521],  # apical
        13: [0.2, 0.2, 0.2],  # axon
        14: [0.2, 0.2, 0.2],  # axon
        15: [0.2, 0.2, 0.2],  # basal
        16: [0.2, 0.2, 0.2],  # basal
    }
    for section_id, section in expected.sections.items():
        section.diameters = diameters[section_id]

    _compare_diameters(expected, small_morph)


def test_build_simple_with_several_apical_sections(config, model_params, small_morph):
    """Test the main build function to diametrize a neuron with a known apical section."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"

    # Duplicate the apical
    for section in small_morph.sections[4].iter():
        if section.id != 4:
            append_fun = small_morph.sections[section.parent.id + 13].append_section
        else:
            append_fun = small_morph.append_root_section
        pts = section.points.copy()
        pts[:, 0] += 1000
        append_fun(PointLevel(pts, section.diameters.copy()), SectionType(3))

    model_params[model][mtype]["apical_point_sec_ids"] = [6, 19]

    build_diameters.build(small_morph, model_params[model][mtype], neurite_types, config[model])

    expected = morphio.mut.Morphology(small_morph)
    diameters = {
        0: [1.39890289, 0.94757283, 0.60740244],  # basal
        1: [1.97971463, 1.43837976, 1.02925932],  # basal
        2: [0.81800926, 0.58568162, 0.54873765],  # basal
        3: [0.73983711, 0.59171486, 0.59403551],  # basal
        4: [3.71459246, 3.34208155, 2.96956968],  # apical
        5: [0.44482154, 0.3843236, 0.37343067],  # apical
        6: [2.96956968, 2.40652895, 1.8434881],  # apical
        7: [1.08870435, 0.77451909, 0.49428192],  # apical
        8: [1.25662541, 0.99705613, 0.84291184],  # apical
        9: [0.63015556, 0.48776197, 0.44297838],  # apical
        10: [0.62148929, 0.47264618, 0.45764357],  # apical
        11: [0.41042924, 0.38227296, 0.41449428],  # apical
        12: [0.40574685, 0.35985672, 0.3745116],  # apical
        13: [0.2, 0.2, 0.2],  # axon
        14: [0.2, 0.2, 0.2],  # axon
        15: [0.2, 0.2, 0.2],  # basal
        16: [0.2, 0.2, 0.2],  # basal
        17: [1.53278005, 1.13844323, 0.88405502],  # apical
        18: [0.66239756, 0.52660555, 0.52660555],  # apical
        19: [0.88405502, 0.70394671, 0.66747582],  # apical
        20: [0.66346329, 0.53578627, 0.53427696],  # apical
        21: [0.59090728, 0.49942747, 0.46960464],  # apical
        22: [0.46960464, 0.46642455, 0.47136903],  # apical
        23: [0.46960464, 0.42222959, 0.41856867],  # apical
        24: [0.38936657, 0.38004285, 0.38004285],  # apical
        25: [0.38936657, 0.45000905, 0.51071393],  # apical
    }
    for section_id, section in expected.sections.items():
        section.diameters = diameters[section_id]

    _compare_diameters(expected, small_morph)


def test_build(config, model_params, neuron, neuron_diametrized):
    """Test the main build function to diametrize a neuron."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"

    build_diameters.build(neuron, model_params[model][mtype], neurite_types, config[model])

    _compare_diameters(neuron_diametrized, neuron)


def test_build_no_seed(config, model_params, neuron, neuron_diametrized):
    """Test main build without seed in config."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"

    seed = config[model].pop("seed")
    np.random.seed(seed)
    build_diameters.build(neuron, model_params[model][mtype], neurite_types, config[model])

    _compare_diameters(neuron_diametrized, neuron)


def test_build_multiple_models_warning(config, model_params, neuron, neuron_diametrized, caplog):
    """Test main build with multiple models in config."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"
    config[model]["models"].append("SKIPPED MODEL")

    caplog.clear()
    caplog.set_level(logging.WARNING)
    build_diameters.build(neuron, model_params[model][mtype], neurite_types, config[model])

    module, level, entry = caplog.record_tuples[0]
    assert module == "diameter_synthesis.build_diameters"
    assert level == 30
    assert entry == "Several models provided, we will only use the first"

    _compare_diameters(neuron_diametrized, neuron)


def test_build_one_sample(test_data_path, config, model_params, neuron):
    """Test main build with 1 sample in config."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"
    config[model]["n_samples"] = 1

    build_diameters.build(neuron, model_params[model][mtype], neurite_types, config[model])

    neuron_diametrized = morphio.mut.Morphology(
        test_data_path / "C030796A-P3_lite_diametrized_1_sample.h5"
    )

    _compare_diameters(neuron_diametrized, neuron)


def test_build_no_sequential(test_data_path, config, model_params, neuron):
    """Test main build with no sequential in config."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "astrocyte"

    config[model] = config["generic"]
    config[model]["models"] = [model]
    config[model]["n_samples"] = 1
    model_params[model] = model_params["generic"]
    model_params[model][mtype]["sibling_ratios"]["apical"]["sequential"] = None

    build_diameters.build(neuron, model_params[model][mtype], neurite_types, config[model])

    neuron_diametrized = morphio.mut.Morphology(
        test_data_path / "C030796A-P3_lite_diametrized_astrocyte.h5"
    )

    _compare_diameters(neuron_diametrized, neuron)


def test_build_small_n_tries_warning(config, model_params, neuron, caplog):
    """Test main build with trunk_max_tries = 1."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"
    config[model]["trunk_max_tries"] = 1

    caplog.clear()
    caplog.set_level(logging.WARNING)
    build_diameters.build(neuron, model_params[model][mtype], neurite_types, config[model])

    assert len(caplog.record_tuples) == 8
    neurite_types = [
        "apical",
        "apical",
        "basal",
        "apical",
        "apical",
        "apical",
        "basal",
        "apical",
    ]
    for i in zip(neurite_types, caplog.record_tuples):
        neurite_type, (module, level, entry) = i
        assert module == "diameter_synthesis.build_diameters"
        assert level == 30
        assert entry == f"max tries attained with {neurite_type}"


def test_build_small_trunk_diam_warning(config, model_params, neuron, caplog):
    """Test trunk diam warning in main build."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"

    TRUNK_FRAC_DECREASE = build_diameters.TRUNK_FRAC_DECREASE
    try:
        build_diameters.TRUNK_FRAC_DECREASE = 999
        caplog.clear()
        caplog.set_level(logging.WARNING)
        build_diameters.build(neuron, model_params[model][mtype], neurite_types, config[model])
    finally:
        build_diameters.TRUNK_FRAC_DECREASE = TRUNK_FRAC_DECREASE

    assert len(caplog.record_tuples) == 5
    for i in caplog.record_tuples:
        module, level, entry = i
        assert module == "diameter_synthesis.build_diameters"
        assert level == 30
        assert entry == "sampled trunk diameter < 0.01, so use 1 instead"


def test_select_model():
    """Test the _select_model function."""
    # pylint: disable=protected-access
    # pylint: disable=no-member
    f_generic = build_diameters._select_model("generic")
    assert f_generic.args == (
        {
            "mode_sibling": "threshold",
            "mode_diameter_power_relation": "threshold",
            "with_asymmetry": True,
            "reduction_factor_max": 1.0,
        },
    )

    f_astrocyte = build_diameters._select_model("astrocyte")
    assert f_astrocyte.args == (
        {
            "mode_sibling": "generic",
            "mode_diameter_power_relation": "generic",
            "with_asymmetry": True,
            "reduction_factor_max": 3.0,
        },
    )

    with pytest.raises(DiameterSynthesisError):
        build_diameters._select_model("UNKNOWN")


def test_sample_sibling_ratio(model_params):
    """Test the _sample_sibling_ratio function."""
    # pylint: disable=protected-access
    mtype = "L5_TPC:A"
    model = "generic"
    d_generic = build_diameters._sample_sibling_ratio(
        model_params[model][mtype], "basal", False, mode="generic"
    )
    assert d_generic == pytest.approx(0.8253851329090136)

    d_threshold = build_diameters._sample_sibling_ratio(
        model_params[model][mtype], "basal", False, mode="threshold"
    )
    assert d_threshold == pytest.approx(0.7244487906052952)

    with pytest.raises(DiameterSynthesisError):
        build_diameters._sample_sibling_ratio(
            model_params[model][mtype], "basal", False, mode="UNKNOWN"
        )


def test_sample_diameter_power_relation(model_params):
    """Test the _sample_diameter_power_relation function."""
    # pylint: disable=protected-access
    mtype = "L5_TPC:A"
    model = "generic"
    d_generic = build_diameters._sample_diameter_power_relation(
        model_params[model][mtype], "basal", False, mode="generic"
    )
    assert d_generic == pytest.approx(4.942498757112429)

    d_threshold = build_diameters._sample_diameter_power_relation(
        model_params[model][mtype], "basal", False, mode="threshold"
    )
    assert d_threshold == pytest.approx(5.494060358805385)

    d_threshold = build_diameters._sample_diameter_power_relation(
        model_params[model][mtype], "basal", False, mode="exact"
    )
    assert d_threshold == pytest.approx(1)

    with pytest.raises(DiameterSynthesisError):
        build_diameters._sample_diameter_power_relation(
            model_params[model][mtype], "basal", False, mode="UNKNOWN"
        )


def test_sample_daughter_diameters(neuron, model_params):
    """Test the _sample_diameter_power_relation function."""
    # pylint: disable=protected-access
    mtype = "L5_TPC:A"
    model = "generic"

    param_tree = {
        "mode_sibling": "threshold",
        "mode_diameter_power_relation": "threshold",
        "with_asymmetry": True,
        "reduction_factor_max": 1.0,
        "neurite_type": "basal",
        "asymmetry_threshold": 1.0,
        "trunk_diam": 2.131444808766456,
        "tot_length": 1085.1178037524223,
        "terminal_diam": 0.35602967635703164,
        "major_sections": [],
    }
    d = build_diameters._sample_daughter_diameters(
        neuron.sections[0], model_params[model][mtype], param_tree
    )
    assert d == pytest.approx([0.9373572556376409, 0.7736807430277025])

    param_tree["with_asymmetry"] = False
    d_no_asymmetry = build_diameters._sample_daughter_diameters(
        neuron.sections[0], model_params[model][mtype], param_tree
    )
    assert d_no_asymmetry == pytest.approx([1.120733059071045, 0.8804980886414063])


def test_diametrize_axon(small_morph):
    """Test the axon diametrizer."""
    build_diameters.diametrize_axon(small_morph)

    expected = morphio.mut.Morphology(small_morph)
    diameters = {
        0: [0.2, 0.2, 0.2],  # basal
        1: [0.2, 0.2, 0.2],  # basal
        2: [0.2, 0.2, 0.2],  # basal
        3: [0.2, 0.2, 0.2],  # basal
        4: [0.2, 0.2, 0.2],  # apical
        5: [0.2, 0.2, 0.2],  # apical
        6: [0.2, 0.2, 0.2],  # apical
        7: [0.2, 0.2, 0.2],  # apical
        8: [0.2, 0.2, 0.2],  # apical
        9: [0.2, 0.2, 0.2],  # apical
        10: [0.2, 0.2, 0.2],  # apical
        11: [0.2, 0.2, 0.2],  # apical
        12: [0.2, 0.2, 0.2],  # apical
        13: [0.2, 0.2, 0.9],  # axon
        14: [0.2, 0.2, 0.9],  # axon
        15: [0.376628, 0.326628, 0.276628],  # basal
        16: [0.753256, 0.703256, 0.653256],  # basal
    }
    for section_id, section in expected.sections.items():
        section.diameters = diameters[section_id]

    _compare_diameters(expected, small_morph, rtol=1e-6)



"""Test the plotting module."""
import neurom as nm
import pytest
from diff_pdf_visually import pdfdiff
from matplotlib import pyplot as plt

from diameter_synthesis import plotting
from diameter_synthesis.exception import DiameterSynthesisError


def test_plot_diameter_diff(neuron_diametrized_path, tmpdir, expected_images_path):
    """Test the plot_diameter_diff function."""
    # Test with existing directory and with positive changes

    # Load neuron with NeuroM
    new_neuron = nm.load_morphology(neuron_diametrized_path)

    # Update the neuron
    for sec in new_neuron.sections:
        sec.points[:, nm.COLS.R] *= 20

    # Plot the figure
    plotting.plot_diameter_diff(
        neuron_diametrized_path, new_neuron, ["basal", "apical"], tmpdir, ext=".pdf"
    )

    # Check the figure
    assert pdfdiff(
        expected_images_path / "test_plot_diameter_diff_mult.pdf",
        tmpdir / neuron_diametrized_path.with_suffix(".pdf").name,
    )

    # Test with new directory and with negative changes

    # Load neuron with NeuroM
    new_neuron = nm.load_morphology(neuron_diametrized_path)

    # Update the neuron
    for sec in new_neuron.sections:
        sec.points[:, nm.COLS.R] /= 20

    # Plot the figure
    plotting.plot_diameter_diff(
        neuron_diametrized_path, new_neuron, ["basal", "apical"], tmpdir / "new_dir", ext=".pdf"
    )

    # Check the figure
    assert pdfdiff(
        expected_images_path / "test_plot_diameter_diff_div.pdf",
        tmpdir / "new_dir" / neuron_diametrized_path.with_suffix(".pdf").name,
    )


def test_plot_distribution_fit(model_params, model_data, tmpdir, expected_images_path):
    """Test the plot_distribution_fit function."""
    # Plot the figure
    plotting.plot_distribution_fit(
        model_data["diameter_power_relation"],
        model_params["generic"]["L5_TPC:A"]["diameter_power_relation"],
        ["basal", "apical"],
        fig_name=tmpdir / "test_plot_distribution_fit",
        ext=".pdf",
    )

    # Check the figures
    assert pdfdiff(
        expected_images_path / "test_plot_distribution_fit.pdf",
        tmpdir / "test_plot_distribution_fit.pdf",
    )
    assert pdfdiff(
        expected_images_path / "test_plot_distribution_fit_scatter.pdf",
        tmpdir / "test_plot_distribution_fit_scatter.pdf",
    )


def test_plot_cumulative_distribution(
    single_pop, single_pop_diametrized, tmpdir, expected_images_path
):
    """Test the plot_cumulative_distribution function."""
    # Plot the figure
    plotting.plot_cumulative_distribution(
        single_pop,
        single_pop_diametrized,
        "segment_radial_distances",
        "segment_volumes",
        ["basal", "apical"],
    )
    plt.savefig(tmpdir / "test_plot_cumulative_distribution.pdf", bbox_inches="tight")
    plt.close()

    # Check the figures
    assert pdfdiff(
        expected_images_path / "test_plot_cumulative_distribution.pdf",
        tmpdir / "test_plot_cumulative_distribution.pdf",
    )


def test_make_cumulative_figures(single_pop, single_pop_diametrized, tmpdir, expected_images_path):
    """Test the make_cumulative_figures function."""
    # Plot the figures
    plotting.make_cumulative_figures(
        single_pop,
        single_pop_diametrized,
        "segment_radial_distances",
        "segment_volumes",
        ["basal", "apical"],
        tmpdir,
        individual=True,
        figname_prefix="with_individual_",
        ext=".pdf",
    )

    # Check the figures
    images_path = expected_images_path / "test_make_cumulative_figures"
    assert pdfdiff(
        images_path / "cumulative_segment_radial_distances_volumes.pdf",
        tmpdir / "with_individual_cumulative_segment_radial_distances_volumes.pdf",
    )
    assert pdfdiff(
        images_path
        / "cumulative_segment_radial_distances_volumes_individual"
        / "0_cumulative_segment_radial_distances_volumes.pdf",
        tmpdir
        / "with_individual_cumulative_segment_radial_distances_volumes_individual"
        / "0_with_individual_cumulative_segment_radial_distances_volumes_C030796A-P3_lite.h5.pdf",
    )


def test_cumulative_analysis(
    single_pop_neurondb,
    tmpdir,
    expected_images_path,
    single_pop_data_dir,
    single_pop_diametrized_data_dir,
):
    """Test the cumulative_analysis function."""
    # Plot the figures
    plotting.cumulative_analysis(
        single_pop_data_dir,
        single_pop_diametrized_data_dir,
        tmpdir / "analysis",
        True,
        single_pop_data_dir / "neurondb.dat",
        ["basal", "apical"],
        ext=".pdf",
    )

    # Check the figures
    images_path = expected_images_path / "test_cumulative_analysis" / "analysis"
    assert pdfdiff(
        images_path / "L5_TPC:Acumulative_section_path_distances_areas.pdf",
        tmpdir / "analysis" / "L5_TPC:Acumulative_section_path_distances_areas.pdf",
    )
    assert pdfdiff(
        images_path / "L5_TPC:Acumulative_section_path_distances_volumes.pdf",
        tmpdir / "analysis" / "L5_TPC:Acumulative_section_path_distances_volumes.pdf",
    )
    assert pdfdiff(
        images_path
        / "L5_TPC:Acumulative_section_path_distances_areas_individual"
        / "0_L5_TPC:Acumulative_section_path_distances_areas_C030796A-P3_lite.pdf",
        tmpdir
        / "analysis"
        / "L5_TPC:Acumulative_section_path_distances_areas_individual"
        / "0_L5_TPC:Acumulative_section_path_distances_areas_C030796A-P3_lite.h5.pdf",
    )
    assert pdfdiff(
        images_path
        / "L5_TPC:Acumulative_section_path_distances_volumes_individual"
        / "0_L5_TPC:Acumulative_section_path_distances_volumes_C030796A-P3_lite.pdf",
        tmpdir
        / "analysis"
        / "L5_TPC:Acumulative_section_path_distances_volumes_individual"
        / "0_L5_TPC:Acumulative_section_path_distances_volumes_C030796A-P3_lite.h5.pdf",
    )

    # Test with unconsistent neurondb
    single_pop_neurondb.loc[0, "morphology"] = "UNKNOWN_MORPHOLOGY"
    single_pop_neurondb.to_csv(
        single_pop_data_dir / "neurondb.dat", sep=" ", header=False, index=False
    )
    with pytest.raises(DiameterSynthesisError):
        plotting.cumulative_analysis(
            single_pop_data_dir,
            single_pop_diametrized_data_dir,
            tmpdir / "analysis",
            True,
            single_pop_data_dir / "neurondb.dat",
            ["basal", "apical"],
            ext=".pdf",
        )


def test_violin_analysis(
    tmpdir, expected_images_path, single_pop_data_dir, single_pop_diametrized_data_dir
):
    """Test the violin_analysis function."""
    # Plot the figures
    plotting.violin_analysis(
        single_pop_data_dir,
        single_pop_diametrized_data_dir,
        tmpdir / "analysis",
        single_pop_data_dir / "neurondb.dat",
    )

    # Check the figures
    assert pdfdiff(
        expected_images_path / "test_violin_analysis.pdf",
        tmpdir / "analysis" / "morphometrics.pdf",
    )



"""Build neurite diameters from a pre-generated model. TO REVIEW!."""
import logging
from collections import deque
from copy import copy
from copy import deepcopy
from functools import partial

import numpy as np
from morphio import IterType
from morphio import SectionType
from numpy.polynomial import polynomial

from diameter_synthesis import morph_functions
from diameter_synthesis import utils
from diameter_synthesis.distribution_fitting import sample_distribution
from diameter_synthesis.exception import DiameterSynthesisError

TRUNK_FRAC_DECREASE = 0.1
N_TRIES_BEFORE_REDUC = 5
L = logging.getLogger(__name__)

STR_TO_TYPES = {
    "basal": SectionType.basal_dendrite,
    "apical": SectionType.apical_dendrite,
    "axon": SectionType.axon,
}

TYPES_TO_STR = {
    SectionType.basal_dendrite: "basal",
    SectionType.apical_dendrite: "apical",
    SectionType.axon: "axon",
}


def _reset_caches():
    """Reset the cached functions."""
    morph_functions.sec_length.cache_clear()
    morph_functions.partition_asymmetry_length.cache_clear()
    morph_functions.lengths_from_origin.cache_clear()
    morph_functions.n_children_downstream.cache_clear()


def _get_neurites(neuron, neurite_type):
    """Get a list of neurites to diametrize.

    Args:
        neuron (morphio.mu.Morphology): a neuron
        neurite_type (morphio.SectionType): the neurite type to consider

    Returns:
        list: list of neurites to consider
    """
    return [
        list(neurite.iter()) for neurite in neuron.root_sections if neurite.type == neurite_type
    ]


def _sample_sibling_ratio(
    params, neurite_type, apply_asymmetry=False, mode="generic", rng=np.random
):
    """Sample a sibling ratio from distribution.

    Args:
        params (dict): model parameters
        neurite_type (str): the neurite type to consider
        apply_asymmetry (bool): asymmetry of current branching point
        mode (str): to use or not the asymmetry_threshold

    Returns:
        float: sibling ratio
    """
    if mode == "generic":
        return sample_distribution(params["sibling_ratios"][neurite_type], rng=rng)
    if mode == "threshold":
        if apply_asymmetry:
            return 0.0
        return sample_distribution(params["sibling_ratios"][neurite_type], rng=rng)
    # This case should never happen since the mode is already checked in `_select_model`
    raise DiameterSynthesisError(f"mode not understood {mode}")


def _sample_diameter_power_relation(
    params, neurite_type, apply_asymmetry=False, mode="generic", rng=np.random
):
    """Sample a diameter power relation from distribution.

    Args:
        params (dict): model parameters
        neurite_type (str): the neurite type to consider
        apply_asymmetry (bool): asymmetry of current branching point
        mode (str): to use or not the asymmetry_threshold

    Returns:
        float: diameter power relation
    """
    if mode == "generic":
        return sample_distribution(params["diameter_power_relation"][neurite_type], rng=rng)
    if mode == "threshold":
        if apply_asymmetry:
            return 1.0
        return sample_distribution(params["diameter_power_relation"][neurite_type], rng=rng)
    if mode == "exact":
        # This case should never happen since this mode is not known by `_select_model`
        return 1.0
    # This case should never happen since the mode is already checked in `_select_model`
    raise DiameterSynthesisError(f"mode not understood {mode}")


def _sample_trunk_diameter(params, neurite_type, rng=np.random):
    """Sample a trunk diameter from distribution.

    Args:
        params (dict): model parameters
        neurite_type (str): the neurite type to consider

    Returns:
        float: trunk diameter
    """
    return sample_distribution(params["trunk_diameters"][neurite_type], rng=rng)


def _sample_terminal_diameter(params, neurite_type, rng=np.random):
    """Sample a terminal diameter.

    Args:
        params (dict): model parameters
        neurite_type (str): the neurite type to consider

    Returns:
        float: terminal diameter
    """
    return sample_distribution(params["terminal_diameters"][neurite_type], rng=rng)


def _sample_taper(params, neurite_type, rng=np.random):
    """Sample a taper rate from distributions.

    Args:
        params (dict): model parameters
        neurite_type (str): the neurite type to consider

    Returns:
        float: taper rate
    """
    return sample_distribution(params["tapers"][neurite_type], rng=rng)


def _sample_daughter_diameters(section, params, params_tree, rng=np.random):
    """Compute the daughter diameters of the current section.

    Args:
        section (morphio  section): section to consider
        params (dict): model parameters
        params_tree (dict): specific parameters of the current tree

    Returns:
       list: list of daughter diameters
    """
    # pylint: disable=too-many-locals
    major_sections = params_tree["major_sections"]

    apply_asymmetry = section.id in major_sections

    reduction_factor = params_tree["reduction_factor_max"] + 1.0
    # try until we get a reduction of diameter in the branching
    while reduction_factor > params_tree["reduction_factor_max"]:

        sibling_ratio = _sample_sibling_ratio(
            params,
            params_tree["neurite_type"],
            apply_asymmetry=apply_asymmetry,
            mode=params_tree["mode_sibling"],
            rng=rng,
        )

        diameter_power_relation = _sample_diameter_power_relation(
            params,
            params_tree["neurite_type"],
            apply_asymmetry=apply_asymmetry,
            mode=params_tree["mode_diameter_power_relation"],
            rng=rng,
        )

        reduction_factor = morph_functions.diameter_power_relation_factor(
            diameter_power_relation, sibling_ratio
        )

    diam_0 = section.diameters[-1]
    terminal_diam = min(diam_0, params_tree["terminal_diam"])

    diam_1 = reduction_factor * diam_0
    diam_2 = sibling_ratio * diam_1

    diam_1 = max(diam_1, terminal_diam)
    diam_2 = max(diam_2, terminal_diam)

    diams = [diam_1] + (len(section.children) - 1) * [diam_2]

    if params_tree.get("with_asymmetry", False):
        # This case should always happen since the `with_asymmetry` attribute is always set to True
        # in `_select_model`

        # returns child diameters sorted by child length (major/secondary for apical tree)
        child_not_in_major = [child.id not in major_sections for child in section.children]
        if False in child_not_in_major:
            child_sort = np.argsort(child_not_in_major)
            return list(np.array(diams)[child_sort])

    # At the moment we don't have enough information to do better than a random choice in this case
    rng.shuffle(diams)
    return diams


def _diametrize_section(section, initial_diam, taper, min_diam=0.07, max_diam=10.0):
    """Diameterize a section.

    Args:
        section (morphio section): current section
        initial_diam (float): initial diameter
        taper (float): taper rate
        min_diam (flaot): minimum diameter
        max_diam (float): maximum diameter
    """
    diams = polynomial.polyval(morph_functions.lengths_from_origin(section), [initial_diam, taper])
    section.diameters = np.clip(diams, min_diam, max_diam)


def _diametrize_tree(neurite, params, params_tree, rng=np.random):
    """Diametrize a tree, or neurite.

    Args:
        neurite (morphio neurite): current neurite
        params (dict): model parameters
        params_tree (dict): specific parameters of the current tree

    Returns:
        bool: True is all terminal diameters are small enough, False otherwise
    """
    params_tree["tot_length"] = morph_functions.get_total_length(neurite)
    max_diam = params["terminal_diameters"][params_tree["neurite_type"]]["params"]["max"]
    wrong_tips = False
    active = deque([neurite[0]])
    while active:
        section = active.popleft()

        if section.is_root:
            init_diam = params_tree["trunk_diam"]
        else:
            init_diam = section.diameters[0]

        taper = _sample_taper(params, params_tree["neurite_type"], rng=rng)

        params_tree["terminal_diam"] = min(
            init_diam, _sample_terminal_diameter(params, params_tree["neurite_type"], rng=rng)
        )

        _diametrize_section(
            section,
            init_diam,
            taper=taper,
            min_diam=params_tree["terminal_diam"],
            max_diam=params_tree["trunk_diam"],
        )

        if len(section.children) > 0:
            diams = _sample_daughter_diameters(section, params, params_tree, rng=rng)

            for i, child in enumerate(section.children):
                utils.redefine_diameter_section(child, 0, diams[i])
                active.append(child)

        # if we are at a tip, check if tip diameters are small enough
        elif section.diameters[-1] > max_diam:
            wrong_tips = True

    return wrong_tips


def _diametrize_neuron(params_tree, neuron, params, neurite_types, config, rng=np.random):
    """Diametrize a neuron.

    Args:
        params_tree (dict): specific parameters of the current tree
        neuron (morphio.mut.Morphology): neuron to diametrize
        params (dict): model parameters
        neurite_types (str or morphio.SectionType): the neurite type to consider
        config (dict): general configuration parameters
        rng: the random number generator to use
    """
    # pylint: disable=too-many-locals, too-many-branches
    major_sections = set()
    if params_tree["with_asymmetry"]:
        # Get sections on the major branch
        for apical_section in params.get("apical_point_sec_ids", []):
            for sec in neuron.sections[apical_section].iter(IterType.upstream):
                major_sections.add(sec.id)
    params_tree["major_sections"] = major_sections

    for neurite_type in neurite_types:
        if isinstance(neurite_type, str):
            morphio_neurite_type = STR_TO_TYPES[neurite_type]
        else:
            morphio_neurite_type = neurite_type
            neurite_type = TYPES_TO_STR[neurite_type]

        params_tree["neurite_type"] = neurite_type

        for neurite in _get_neurites(neuron, morphio_neurite_type):
            wrong_tips = True
            n_tries = 0
            trunk_diam_frac = 1.0
            n_tries_step = 1

            _params = deepcopy(params) if "taper_increase" in params_tree else params
            while wrong_tips:
                if "taper_increase" in params_tree:
                    _params["tapers"][neurite_type]["params"]["scale"] *= params_tree[
                        "taper_increase"
                    ]
                    params_tree["trunk_diam"] = neurite[0].diameters[0]
                else:
                    trunk_diam = trunk_diam_frac * _sample_trunk_diameter(
                        _params, neurite_type, rng=rng
                    )
                    if trunk_diam < 0.01:
                        trunk_diam = 1.0
                        L.warning("sampled trunk diameter < 0.01, so use 1 instead")
                    params_tree["trunk_diam"] = trunk_diam

                wrong_tips = _diametrize_tree(neurite, _params, params_tree, rng=rng)

                # if we can't get a good model, reduce the trunk diameter progressively
                if n_tries > N_TRIES_BEFORE_REDUC * n_tries_step:
                    trunk_diam_frac -= TRUNK_FRAC_DECREASE
                    n_tries_step += 1
                if n_tries > config["trunk_max_tries"]:
                    L.warning("max tries attained with %s", neurite_type)
                    wrong_tips = False
                n_tries += 1


def _select_model(model):
    """Select a diametrized model to use.

    Args:
        model (str): model name

    Returns:
        function: diamtrizer with specific params_tree
    """
    if model == "generic":
        params_tree = {}
        params_tree["mode_sibling"] = "threshold"
        params_tree["mode_diameter_power_relation"] = "threshold"
        params_tree["with_asymmetry"] = True
        params_tree["reduction_factor_max"] = 1.0
    elif model == "astrocyte":
        params_tree = {}
        params_tree["mode_sibling"] = "generic"
        params_tree["mode_diameter_power_relation"] = "generic"
        params_tree["with_asymmetry"] = True
        params_tree["reduction_factor_max"] = 3.0
    elif model == "neurite_based":
        params_tree = {}
        params_tree["mode_sibling"] = "threshold"
        params_tree["mode_diameter_power_relation"] = "threshold"
        params_tree["with_asymmetry"] = True
        params_tree["reduction_factor_max"] = 1.0
        params_tree["taper_increase"] = 1.02

    else:
        raise DiameterSynthesisError(f"Unknown diameter model: {model}")

    return partial(_diametrize_neuron, params_tree)


def build(neuron, model_params, neurite_types, config, rng=np.random):
    """Builder function for generating diameters of a neuron from the a diameter models.

    Args:
        neuron (morphio.mut.Morphology): neuron to diametrize
        model_params (dict): model parameters
        neurite_types (str): the neurite type to consider
        config (dict): general configuration parameters
        rng: random number generator to use
    """
    if "seed" in config:
        np.random.seed(config["seed"])
    _reset_caches()

    if len(config["models"]) > 1:
        L.warning("Several models provided, we will only use the first")
    diameter_generator = _select_model(config["models"][0])

    diameter_generator(neuron, model_params, neurite_types, config, rng=rng)
    if config["n_samples"] > 1:
        diameters = utils.get_all_diameters(neuron)
        for _ in range(config["n_samples"] - 1):
            diameter_generator(neuron, model_params, neurite_types, config, rng=rng)
            for i, new_diams in enumerate(utils.get_all_diameters(neuron)):
                diameters[i] += new_diams
        for i, _ in enumerate(diameters):
            diameters[i] /= config["n_samples"]
        utils.set_all_diameters(neuron, diameters)


def _save_first_diams(morphology, length):
    """Save diameters in a dict up to length."""
    diams = {}
    for root_section in morphology.root_sections:
        if root_section.type == SectionType.axon:
            dist = 0
            prev_point = root_section.points[0]
            for section in root_section.iter():
                _diams = copy(section.diameters)
                for point in section.points:
                    dist += np.linalg.norm(point - prev_point)
                    prev_point = copy(point)
                    if dist >= length:
                        break
                diams[section.id] = _diams
                if dist >= length:
                    break

    return diams


def _set_first_diams(morphology, diams, length):
    """Set diameters from a dict up to length."""
    for root_section in morphology.root_sections:
        if root_section.type == SectionType.axon:
            dist = 0
            prev_point = root_section.points[0]
            for section in root_section.iter():
                current_diams = copy(section.diameters)
                old_diams = diams[section.id]
                for i, point in enumerate(section.points):
                    dist += np.linalg.norm(point - prev_point)
                    prev_point = copy(point)
                    current_diams[i] = old_diams[i]
                    if dist >= length:
                        break
                section.diameters = current_diams
                if dist >= length:
                    break


def diametrize_axon(
    morphology,
    main_diameter=1.0,
    colateral_diameter=0.1,
    main_taper=-0.0005,
    axon_point_isec=None,
    ais_length=60,
    rng=np.random,
):
    """Diametrize axon in place without learning from reconstructed axons.

    The main axon branch (from soma to axon point) will have a tapered diameter, and the
    colaterals a constant diameter.

    If an axon point is not provided, and main_diameter > colateral_diameter, the diameters will
    decrease with taper and bifurcations, with hardcoded parameters sibling_ratio = 0.5 and
    diameter_power_relation = 0.5.

    Args:
        morphology (morphio.mut.Morphology): morpholoty to diametrize
        main_diameter (float): diameter of main axon branch (from soma to axon_point_isec)
        colateral_diameter (float): diameter of colateral branches
        main_taper (float): taper rate of main branch (set to 0 for no taper, should be negative)
        axon_point_isec (int): morphio section id of axon point (see morph_tool.axon_point module)
        ais_length (float): length of ais for which we keep original diameters
        rng: random number generator to use
    """
    model_params = {
        "trunk_diameters": {
            "axon": {
                "distribution": "constant",
                "params": {"value": main_diameter},
            }
        },
        "terminal_diameters": {
            "axon": {
                "distribution": "constant",
                "params": {"value": colateral_diameter, "max": main_diameter},
            }
        },
        "tapers": {
            "axon": {
                "distribution": "constant",
                "params": {"value": main_taper},
            }
        },
        "sibling_ratios": {
            "axon": {
                "distribution": "constant",
                "params": {"value": 0.5},
            }
        },
        "diameter_power_relation": {
            "axon": {
                "distribution": "constant",
                "params": {"value": 5.0},
            }
        },
    }
    if axon_point_isec:
        model_params["apical_point_sec_ids"] = [axon_point_isec]

    config = {"models": ["generic"], "trunk_max_tries": 1, "n_samples": 1}

    diams = _save_first_diams(morphology, ais_length)
    build(morphology, model_params, neurite_types=["axon"], config=config, rng=rng)
    _set_first_diams(morphology, diams, ais_length)



"""Utils functions. TO REVIEW!."""
import json
import logging
import os
from collections import defaultdict
from pathlib import Path
from xml.etree import ElementTree as ET

import numpy as np
import pandas as pd
from neurom import COLS

from diameter_synthesis.exception import DiameterSynthesisError

L = logging.getLogger(__name__)


def _create_morphologies_dict_dat(morph_path, mtypes_file="neurondb.dat"):
    """Create dict to load the morphologies from a directory, with dat file.

    Args:
        morph_path (str): path to morphologies
        mtype_file (str): path to dat file

    Returns:
        dict: dictionary of morphologies keyed by mtypes
    """
    # pylint wrongly determines `morph_name` as TextFileReader
    # pylint: disable=no-member
    morph_name = pd.read_csv(mtypes_file, sep=r"\s+", header=None)
    name_dict = defaultdict(list)
    if not morph_name.empty:
        first_name = morph_name.loc[0, 0]  # pylint: disable=no-member
        file_list = Path(morph_path).glob(first_name + "*")
        try:
            ext = next(file_list).suffix
        except StopIteration as e:
            raise DiameterSynthesisError(f"Could not find a file starting with {first_name}") from e
    for morph in morph_name.values:  # pylint: disable=no-member
        name_dict[morph[2]] += [Path(morph_path) / (morph[0] + ext)]
    return name_dict


def _create_morphologies_dict_folder(morph_path):
    """Create dict to load the morphologies from a directory, from folders.

    Args:
        morph_path (str): path to morphologies

    Returns:
        dict: dictionary of morphologies keyed by mtypes
    """
    name_dict = defaultdict(list)
    for mtype in Path(morph_path).iterdir():
        for fname in mtype.iterdir():
            if fname.suffix in [".h5", ".asc", ".swc"]:
                name_dict[mtype.parts[-1]] += [fname]
    return name_dict


def _create_morphologies_dict_all(morph_path):
    """Create dict to load the morphologies from a directory, all together.

    Args:
        morph_path (str): path to morphologies

    Returns:
        dict: dictionary of morphologies with single key
    """
    name_dict = {"generic_type": []}
    for fname in Path(morph_path).iterdir():
        if fname.suffix in [".h5", ".asc", ".swc"] and fname.exists():
            name_dict["generic_type"] += [fname]

    return name_dict


def create_morphologies_dict(morph_path, mtypes_file=None):
    """Create dict to load the morphologies from a directory, by mtype.

     Args:
        morph_path (str): path to morphologies
        mtype_file (str): path to dat file

    Returns:
        dict: dictionary of morphologies with single key
    """
    if mtypes_file is None and next(Path(morph_path).iterdir()).is_dir():
        L.info("found folder structure per mtype")
        return _create_morphologies_dict_folder(morph_path)
    if mtypes_file is not None:
        # mtype_path = Path(morph_path) / mtypes_file
        mtype_path = Path(mtypes_file)
        if mtype_path.exists():
            if mtype_path.suffix == ".dat":
                L.info("found dat file")
                return _create_morphologies_dict_dat(morph_path, mtypes_file=mtypes_file)
            raise DiameterSynthesisError(
                f"neurondb file format {mtype_path.suffix} not implemented"
            )

    L.info("use all files as single mtype")
    return _create_morphologies_dict_all(morph_path)


def _set_diameters(section, diameters):
    """Set diameters (neurom 2.0.0 only).

    Args:
        section (neurom 2.0.0 section): section to diametrize
        diameters (list/ndarray): diameters
    """
    section.morphio_section.diameters = diameters


def _get_mean_diameter(section):
    """Section mean diameter.

    It is obtained by averaging the segment truncated cone
    diameters and weighting them by their length. (neurom only)

    Args:
        section (neurom section): section to consider

    Returns:
        float: mean diameter of the section
    """
    segment_lengths = np.linalg.norm(
        section.points[1:, COLS.XYZ] - section.points[:-1, COLS.XYZ], axis=1
    )
    segment_mean_diams = section.points[1:, COLS.R] + section.points[:-1, COLS.R]
    return np.sum(segment_mean_diams * segment_lengths) / segment_lengths.sum()


def get_all_diameters(neuron):
    """Get all neuron diameters (morphio only).

    Args:
        neuron (morphio.mut.Morphology): neuron to consider

    Returns:
        list: all diameters
    """
    return [np.array(section.diameters) for section in neuron.iter()]


def set_all_diameters(neuron, all_diameters):
    """Set all neuron diameters (morphio only).

    Args:
        neuron (morphio.mut.Morophology): neuron to consider
        all_diameters (list): list of section diameters
    """
    for diameters, section in zip(all_diameters, neuron.iter()):
        section.diameters = diameters


def _get_diameters(section):
    """Get diameters (neurom only).

    Args:
        section (neurom section): section to consider

    Return:
        list: diameters of section
    """
    return section.points[:, COLS.R] * 2.0


def redefine_diameter_section(section, diam_ind, diam_new):
    """Replace given diameters at indices diam_ind with values diam_new (morphio only).

    Args:
        section (neurom section): section to consider
        diam_ind (list): indices of diameters, or points on the section
        diam_new (list): corresponding diameters
    """
    diameters = section.diameters
    diameters[diam_ind] = diam_new
    section.diameters = diameters


#######################################################################
# old parser functions (will be removed or integrated at some points) #
#######################################################################
def _create_morphologies_dict_json(
    morph_path,
    mtypes_file="neuronDB.xml",
    prefix="",
):
    """Create dict to load the morphologies from a directory, with json."""
    with open(mtypes_file, "r") as filename:
        morph_name = json.load(filename)

    name_dict = {}
    for fname in os.listdir(morph_path):
        filepath = Path(morph_path) / fname
        if fname.endswith((".h5", ".asc", ".swc")) and filepath.exists():
            mtype = morph_name[Path(fname).stem]
            if mtype in name_dict:
                name_dict[mtype] += [prefix + fname]
            else:
                name_dict[mtype] = [prefix + fname]

    return name_dict


def _create_morphologies_dict_xml(
    morph_path,
    mtypes_file="neuronDB.xml",
    ext=".asc",
    prefix="",
):
    """Create dict to load the morphologies from a directory, from xml."""
    filedb = ET.parse(morph_path + mtypes_file)
    root = filedb.findall("listing")[0]
    morphs = root.findall("morphology")

    name_dict = {}
    if len(morphs) > 0:
        for morph in morphs:
            try:
                # Define mtypes
                mtype = morph.find("mtype").text
                # Define subtypes (if they exist)
                if morph.find("msubtype").text:
                    mtype = mtype + ":" + morph.find("msubtype").text

                if mtype not in name_dict:
                    name_dict[mtype] = [prefix + morph.find("name").text + ext]
                elif mtype in name_dict:
                    name_dict[mtype] += [prefix + morph.find("name").text + ext]

            except Exception as exc:  # pylint: disable=broad-except
                L.exception("Failed to process %s", exc)

    return name_dict



"""Main functions to learn and generate diameters."""
import json
import logging
import multiprocessing
import os
from pathlib import Path

import neurom as nm
import numpy as np
from morphio.mut import Morphology
from tqdm import tqdm

from diameter_synthesis import utils
from diameter_synthesis.build_diameters import build as build_diameters
from diameter_synthesis.build_models import build as build_model
from diameter_synthesis.plotting import plot_distribution_fit

L = logging.getLogger(__name__)
logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))


class NumpyEncoder(json.JSONEncoder):
    """Class to encode numpy arrays."""

    def default(self, o):  # pylint: disable=method-hidden
        """Numpy encoder."""
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, np.floating):
            return float(o)
        if isinstance(o, np.integer):
            return int(o)
        return json.JSONEncoder.default(self, o)


def plot_models(morphologies, config, models_params, models_data, ext="png"):
    """Plot the models."""
    L.info("Plot the fits...")

    if not Path(config["fig_folder"]).exists():
        os.mkdir(config["fig_folder"])

    for mtype in tqdm(morphologies):
        if not (Path(config["fig_folder"]) / mtype).exists():
            os.mkdir(Path(config["fig_folder"]) / mtype)
        for model in config["models"]:
            fit_tpes = models_data[model][mtype]
            model_data = models_data[model][mtype]
            model_param = models_params[model][mtype]

            for fit_tpe in fit_tpes:
                fig_name = Path(config["fig_folder"]) / mtype / fit_tpe
                plot_distribution_fit(
                    model_data[fit_tpe],
                    model_param[fit_tpe],
                    config["neurite_types"],
                    fig_name=fig_name,
                    ext=ext,
                )


def _build_all_models(morphologies, config, plot=False, ext="png"):
    """Build all the models in the list of models."""
    models_params = {}
    models_data = {}
    for model in config["models"]:
        models_params[model] = {}
        models_data[model] = {}
        for mtype in tqdm(morphologies):
            config_model = config.copy()
            config_model["models"] = [model]
            models_params[model][mtype], models_data[model][mtype] = build_model(
                morphologies[mtype], config_model, with_data=True
            )

    if plot:
        plot_models(morphologies, config, models_params, models_data, ext=ext)

    return models_params


def run_models(config_file, plot, ext="png"):
    """Run the model extraction from config file."""
    with open(config_file, "r") as filename:
        config = json.load(filename)

    L.info("Loading morphologies...")
    morphologies_dict = utils.create_morphologies_dict(
        config["morph_path"],
        mtypes_file=config["mtypes_file"],
    )

    morphologies = {
        mtype: [nm.load_morphology(i) for i in morphologies_dict[mtype]]
        for mtype in tqdm(morphologies_dict)
    }

    L.info("Extracting model parameters...")
    models_params = _build_all_models(morphologies, config, plot=plot, ext=ext)

    with open(config["models_params_file"], "w") as json_file:
        json.dump(models_params, json_file, sort_keys=True, indent=4, cls=NumpyEncoder)


class DiameterWorker:
    """Worker for building diameters."""

    def __init__(self, model, models_params, config):
        """Init function."""
        self.model = model
        self.models_params = models_params[model]
        self.config = config

    def __call__(self, neuron_input):
        """Call function."""
        fname = neuron_input[0]
        mtype = neuron_input[1]

        neuron = Morphology(fname)

        build_diameters(
            neuron,
            self.models_params[mtype],
            self.config["neurite_types"],
            self.config,
        )

        if not Path(self.config["new_morph_path"]).exists():
            os.mkdir(self.config["new_morph_path"])

        if fname.parts[-2] == mtype:
            if not (Path(self.config["new_morph_path"]) / fname.parts[-2]).exists():
                os.mkdir(Path(self.config["new_morph_path"]) / fname.parts[-2])
            neuron.write(Path(self.config["new_morph_path"]) / fname.parts[-2] / fname.name)
        neuron.write(Path(self.config["new_morph_path"]) / fname.name)


def run_diameters(config_file, models_params_file):
    """Build new diameters from config file and diameter model."""
    with open(config_file, "r") as filename:
        config = json.load(filename)

    with open(models_params_file, "r") as filename:
        models_params = json.load(filename)

    for model in config["models"]:
        L.info("Generating diameter with model %s", model)

        morphologies_dict = utils.create_morphologies_dict(
            config["morph_path"],
            mtypes_file=config["mtypes_file"],
        )

        worker = DiameterWorker(model, models_params, config)

        all_neurons = [
            [neuron, mtype] for mtype in morphologies_dict for neuron in morphologies_dict[mtype]
        ]

        pool = multiprocessing.Pool(config["n_cpu"])  # pylint: disable=consider-using-with
        list(tqdm(pool.imap(worker, all_neurons), total=len(all_neurons)))
        pool.close()
        pool.join()


def diametrize_single_neuron(neuron, config=None, apical_point_sec_ids=None):
    """Diametrize single neuron by learning diameter model from it.

    Args:
        neuron (mophio.mut.Morphology): neuron to consider
        config (dict): dict with entry 'model' and 'diameters' with corresponding dicts, if None,
            default dict will be used
        apical_point_sec_ids (list): list of apical points if any
    """
    if config is None:
        config = {
            "model": {
                "taper": {"max": 1e-06, "min": -0.1},
                "terminal_threshold": 2.0,
                "models": ["generic"],
                "neurite_types": ["basal", "apical"],
            },
            "diameters": {
                "models": ["generic"],
                "n_samples": 1,
                "neurite_types": ["basal", "apical"],
                "seed": 0,
                "trunk_max_tries": 100,
            },
        }

    model_params = build_model([nm.load_neuron(neuron)], config["model"])
    if apical_point_sec_ids is not None:
        model_params["apical_point_sec_ids"] = apical_point_sec_ids
    build_diameters(neuron, model_params, ["basal", "apical"], config["diameters"])



"""Setup for the diameter-synthesis package."""
import imp
import sys
from pathlib import Path

from setuptools import find_packages
from setuptools import setup

if sys.version_info < (2, 7):
    sys.exit("Sorry, Python < 2.7 is not supported")

VERSION = imp.load_source("", "diameter_synthesis/version.py").__version__

test_reqs = [
    "diff_pdf_visually>=1.5.1",
    "matplotlib<=3.3.4",
    "decorator",
    "matplotlib<3.4",  # PDF rendering is slightly modified with matplotlib >= 3.4
    "mock",
    "pytest",
    "pytest-cov",
    "pytest-html",
    "pytest-xdist",
]

setup(
    name="diameter-synthesis",
    author="BlueBrain Cells",
    version=VERSION,
    description="Diametrize cells",
    long_description=Path("README.rst").read_text(encoding="utf-8"),
    long_description_content_type="text/x-rst",
    license="GNU General Public License v3.0",
    url="https://diameter-synthesis.readthedocs.io",
    project_urls={
        "Tracker": "https://github.com/BlueBrain/diameter-synthesis/issues",
        "Source": "https://github.com/BlueBrain/diameter-synthesis",
    },
    install_requires=[
        "click>=7.0",
        "jsonschema>=3",
        "matplotlib>=2.2.0",
        "morphio>=2.3.4",
        "neurom>=3.0,<4.0",
        "numpy>=1.15.0",
        "pandas>=0.24.0",
        "scipy>=0.13.3",
    ],
    extras_require={
        "plot": "seaborn>=0.11.1",
        "test": test_reqs,
    },
    tests_require=test_reqs,
    entry_points={
        "console_scripts": ["diameter-synthesis=diameter_synthesis.cli:cli"],
    },
    python_requires=">=3.6",
    packages=find_packages(include=["diameter_synthesis*"], exclude=["tests"]),
    include_package_data=True,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)
